# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import os
import json
import numpy as np
from cherab.core.utility import RecursiveDict
from cherab.core.atomic import Element
from .utility import DEFAULT_REPOSITORY_PATH, valid_ionisation, encode_transition

"""
Utilities for managing the local rate repository - PEC section.
"""


def add_pec_excitation_rate(element, ionisation, transition, rate, repository_path=None):
    """
    Adds a single PEC excitation rate to the repository.

    If adding multiple rate, consider using the update_pec_rates() function
    instead. The update function avoid repeatedly opening and closing the rate
    files.

    :param element:
    :param ionisation:
    :param transition:
    :param rate:
    :param repository_path:
    :return:
    """

    update_pec_rates({
        'excitation': {
            element: {
                ionisation: {
                    transition: rate
                }
            }
        }
    }, repository_path)


def add_pec_recombination_rate(element, ionisation, transition, rate, repository_path=None):
    """
    Adds a single PEC recombination rate to the repository.

    If adding multiple rate, consider using the update_pec_rates() function
    instead. The update function avoid repeatedly opening and closing the rate
    files.

    :param element:
    :param ionisation:
    :param transition:
    :param rate:
    :param repository_path:
    :return:
    """

    update_pec_rates({
        'recombination': {
            element: {
                ionisation: {
                    transition: rate
                }
            }
        }
    }, repository_path)


def add_pec_thermalcx_rate(element, ionisation, transition, rate, repository_path=None):
    """
    Adds a single PEC thermalcx rate to the repository.

    If adding multiple rate, consider using the update_pec_rates() function
    instead. The update function avoid repeatedly opening and closing the rate
    files.

    :param element:
    :param ionisation:
    :param transition:
    :param rate:
    :param repository_path:
    :return:
    """

    update_pec_rates({
        'thermalcx': {
            element: {
                ionisation: {
                    transition: rate
                }
            }
        }
    }, repository_path)


def update_pec_rates(rates, repository_path=None):
    """
    PEC rate file structure

    /pec/<class>/<element>/<ionisation>.json
    """

    valid_classes = [
        'excitation',
        'recombination',
        'thermalcx'
    ]

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for cls, elements in rates.items():
        for element, ionisations in elements.items():
            for ionisation, transitions in ionisations.items():

                # sanitise and validate
                cls = cls.lower()
                if cls not in valid_classes:
                    raise ValueError('Unrecognised pec rate class \'{}\'.'.format(cls))

                if not isinstance(element, Element):
                    raise TypeError('The element must be an Element object.')

                if not valid_ionisation(element, ionisation):
                    raise ValueError('Ionisation level is larger than the number of protons in the element.')

                path = os.path.join(repository_path, 'pec/{}/{}/{}.json'.format(cls, element.symbol.lower(), ionisation))

                # read in any existing rates
                try:
                    with open(path, 'r') as f:
                        content = RecursiveDict.from_dict(json.load(f))
                except FileNotFoundError:
                    content = RecursiveDict()

                # add/replace data for a transition
                for transition in transitions:
                    key = encode_transition(transition)
                    data = rates[cls][element][ionisation][transition]

                    # sanitise/validate data
                    data['ne'] = np.array(data['ne'], np.float64)
                    data['te'] = np.array(data['te'], np.float64)
                    data['rate'] = np.array(data['rate'], np.float64)

                    if data['ne'].ndim != 1:
                        raise ValueError('Density array must be a 1D array.')

                    if data['te'].ndim != 1:
                        raise ValueError('Temperature array must be a 1D array.')

                    if (data['ne'].shape[0], data['te'].shape[0]) != data['rate'].shape:
                        raise ValueError('Density, temperature and rate data arrays have inconsistent sizes.')

                    content[key] = {
                        'ne': data['ne'].tolist(),
                        'te': data['te'].tolist(),
                        'rate': data['rate'].tolist()
                    }

                # create directory structure if missing
                directory = os.path.dirname(path)
                if not os.path.isdir(directory):
                    os.makedirs(directory)

                # write new data
                with open(path, 'w') as f:
                    json.dump(content, f, indent=2, sort_keys=True)


def get_pec_excitation_rate(element, ionisation, transition, repository_path=None):
    return _get_pec_rate('excitation', element, ionisation, transition, repository_path)


def get_pec_recombination_rate(element, ionisation, transition, repository_path=None):
    return _get_pec_rate('recombination', element, ionisation, transition, repository_path)


def get_pec_thermalcx_rate(element, ionisation, transition, repository_path=None):
    return _get_pec_rate('thermalcx', element, ionisation, transition, repository_path)


def _get_pec_rate(cls, element, ionisation, transition, repository_path=None):

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH
    path = os.path.join(repository_path, 'pec/{}/{}/{}.json'.format(cls, element.symbol.lower(), ionisation))
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        d = content[encode_transition(transition)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested PEC rate (class={}, element={}, ionisation={}, transition={})'
                           ' is not available.'.format(cls, element.symbol, ionisation, transition))

    # convert to numpy arrays
    d['ne'] = np.array(d['ne'], np.float64)
    d['te'] = np.array(d['te'], np.float64)
    d['rate'] = np.array(d['rate'], np.float64)

    return d





# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

# todo: to be implemented in a future release

# import numpy as np
# from numpy cimport ndarray
# from cherab.core.math.interpolators.interpolators2d cimport Interpolate2DCubic
# from cherab.core.atomic.rates cimport StageResolvedLineRadiation as CoreStageResolvedLineRadiation
#
#
# # todo remove log10 of data
# cdef class StageResolvedRadiation(CoreStageResolvedLineRadiation):
#
#     cdef:
#         readonly bint extrapolate
#         readonly tuple density_range, temperature_range
#         readonly ndarray _electron_density, _electron_temperature, _radiated_power
#         readonly Interpolate2DCubic _power_func
#
#     def __init__(self, element, ionisation, electron_density, electron_temperature, radiated_power, name='', extrapolate=False):
#
#         super().__init__(element, ionisation, name=name)
#
#         self._electron_density = np.array(electron_density, dtype=np.float64)
#         self._electron_temperature = np.array(electron_temperature, dtype=np.float64)
#         self._radiated_power = np.array(radiated_power, dtype=np.float64)
#
#         if not all(j >= 0 for i in self._radiated_power for j in i):
#             raise ValueError("Can't have negative power values!")
#
#         self.density_range = (self._electron_density.min(), self._electron_density.max())
#         self.temperature_range = (self._electron_temperature.min(), self._electron_temperature.max())
#
#         self.extrapolate = extrapolate
#         self._power_func = Interpolate2DCubic(self._electron_density, self._electron_temperature, self._radiated_power,
#                                               extrapolate=extrapolate, extrapolation_type="nearest")
#
#     cdef double evaluate(self, double electron_density, double electron_temperature) except? -1e999:
#         cdef double rate = self._power_func.evaluate(electron_density, electron_temperature)
#         return max(0, rate)
#
#
# cdef class NullStageResolvedLineRadiation(CoreStageResolvedLineRadiation):
#     """
#     A stage resolved line radiation rate that always returns zero.
#     Needed for use cases where the required atomic data is missing.
#     """
#
#     cdef double evaluate(self, double electron_density, double electron_temperature) except? -1e999:
#         return 0.0



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
import numpy as np
from cherab.core.utility import RecursiveDict
from cherab.core.utility.conversion import Cm3ToM3, PerCm3ToPerM3
from .utility import readvalues


def parse_adf12(donor_ion, donor_metastable, receiver_ion, receiver_ionisation, adf_file_path):
    """
    Opens and parses ADAS ADF12 data files.

    :param donor_ion: The donor ion element described by the rate file.
    :param donor_metastable: The donor ion metastable level.
    :param receiver_ion: The receiver ion element described by the rate file.
    :param receiver_ionisation: The receiver ion ionisation level described by the rate file.
    :param adf_file_path: Path to ADF15 file from ADAS root.
    :return: Dictionary containing rates.
    """

    rates = RecursiveDict()

    with open(adf_file_path, 'r') as file:

        rate_count = int(file.readline()[3:5])
        for i in range(rate_count):

            # parse block
            transition, rate = _parse_block(file)

            # add to repository update dictionary, converting density from cm^-3 to m^-3
            rates[donor_ion][receiver_ion][receiver_ionisation][transition][donor_metastable] = {
                'eb': np.array(rate['ENER'], np.float64),
                'ti': np.array(rate['TIEV'], np.float64),
                'ni': PerCm3ToPerM3.to(np.array(rate['DENSI'], np.float64)),
                'z': np.array(rate['ZEFF'], np.float64),
                'b': np.array(rate['BMAG'], np.float64),

                'qeb': Cm3ToM3.to(np.array(rate['QENER'], np.float64)),
                'qti': Cm3ToM3.to(np.array(rate['QTIEV'], np.float64)),
                'qni': Cm3ToM3.to(np.array(rate['QDENSI'], np.float64)),
                'qz': Cm3ToM3.to(np.array(rate['QZEFF'], np.float64)),
                'qb': Cm3ToM3.to(np.array(rate['QBMAG'], np.float64)),

                'ebref': rate['EBREF'],
                'tiref': rate['TIREF'],
                'niref': PerCm3ToPerM3.to(rate['NIREF']),
                'zref': rate['ZEREF'],
                'bref': rate['BREF'],
                'qref': Cm3ToM3.to(rate['QEFREF'])
            }

    return rates


def _parse_block(file):
    """
    Reads and parses an ADF12 rate block from an IO stream.

    :param file: Text stream seeked to the start of the block.
    :return: Tuple containing (transition tuple, rate data dictionary).
    """

    # header
    line = file.readline()
    transition = (int(line[38:40]), int(line[41:43]))

    rate = {}

    # reference value section
    rate['QEFREF'] = readvalues(file, 1, 6)[0]
    ebref, tiref, niref, zeref, bref = readvalues(file, 5, 6)
    rate['EBREF'] = ebref
    rate['TIREF'] = tiref
    rate['NIREF'] = niref
    rate['ZEREF'] = zeref
    rate['BREF'] = bref

    # rate data section
    nbeam, nti, ndi, nze, nb = readvalues(file, 5, 6, type=int)
    rate['ENER'] = readvalues(file, 24, 6)[0:nbeam]
    rate['QENER'] = readvalues(file, 24, 6)[0:nbeam]
    rate['TIEV'] = readvalues(file, 12, 6)[0:nti]
    rate['QTIEV'] = readvalues(file, 12, 6)[0:nti]
    rate['DENSI'] = readvalues(file, 24, 6)[0:ndi]
    rate['QDENSI'] = readvalues(file, 24, 6)[0:ndi]
    rate['ZEFF'] = readvalues(file, 12, 6)[0:nze]
    rate['QZEFF'] = readvalues(file, 12, 6)[0:nze]
    rate['BMAG'] = readvalues(file, 12, 6)[0:nb]
    rate['QBMAG'] = readvalues(file, 12, 6)[0:nb]

    return transition, rate



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
import os
import json
import numpy as np
from cherab.core.utility import RecursiveDict
from cherab.core.atomic import Element
from ..utility import DEFAULT_REPOSITORY_PATH, valid_ionisation, encode_transition

"""
Utilities for managing the local rate repository - beam cx rate section.
"""


def add_beam_cx_rate(donor_ion, donor_metastable, receiver_ion, receiver_ionisation, transition, rate, repository_path=None):
    """
    Adds a single beam CX rate to the repository.

    If adding multiple rate, consider using the update_beam_cx_rates() function
    instead. The update function avoid repeatedly opening and closing the rate
    files.

    :param donor_ion:
    :param donor_metastable:
    :param receiver_ion:
    :param receiver_ionisation:
    :param rate:
    :param repository_path:
    :return:
    """

    update_beam_cx_rates({
        donor_ion: {
            receiver_ion: {
                receiver_ionisation: {
                    transition: {
                        donor_metastable: rate
                    }
                }
            }
        }
    }, repository_path)


def update_beam_cx_rates(rates, repository_path=None):
    # organisation in repository:
    #   beam/cx/donor_ion/receiver_ion/receiver_ionisation.json
    # inside json file:
    #   transition: [list of donor_metastables with rates]

    def sanitise_and_validate(data, x_key, x_name, y_key, y_name):
        """
        Sanitises and validates pairs of rate data arrays.

        Converts arrays to numpy arrays and check that the dimensions of the
        supplied data are consistent.

        Arrays are converted in place, the supplied dictionary is modified.

        :param data: Rate data dictionary.
        :param x_key: Key of independent variable data.
        :param x_name: Name of data array for error reporting.
        :param y_key: Key of dependent variable data.
        :param y_name: Name of data array for error reporting.
        """

        # convert to numpy arrays
        data[x_key] = np.array(data[x_key], np.float64)
        data[y_key] = np.array(data[y_key], np.float64)

        # check dimensions
        if data[x_key].ndim != 1:
            raise ValueError('The {} array must be a 1D array.'.format(x_name))

        if data[y_key].ndim != 1:
            raise ValueError('The {} array must be a 1D array.'.format(y_name))

        if data[x_key].shape != data[y_key].shape:
            raise ValueError('The {} and {} arrays have inconsistent lengths.'.format(x_name, y_name))

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for donor, receivers in rates.items():
        for receiver, ionisations in receivers.items():
            for ionisation, transitions in ionisations.items():

                # sanitise and validate
                if not isinstance(donor, Element):
                    raise TypeError('The element must be an Element object.')

                if not isinstance(receiver, Element):
                    raise TypeError('The element must be an Element object.')

                if not valid_ionisation(receiver, ionisation):
                    raise ValueError('Ionisation level is larger than the number of protons in the element.')

                path = os.path.join(repository_path, 'beam/cx/{}/{}/{}.json'.format(donor.symbol.lower(), receiver.symbol.lower(), ionisation))

                # read in any existing rates
                try:
                    with open(path, 'r') as f:
                        content = RecursiveDict.from_dict(json.load(f))
                except FileNotFoundError:
                    content = RecursiveDict()

                # json keys are strings, must convert metastable key to integer
                for transition, metastables in content.items():
                    content[transition] = RecursiveDict({int(metastable): rate for metastable, rate in metastables.items()})

                # update content
                for transition, metastables in transitions.items():

                    # add/replace data for each metastable
                    transition_key = encode_transition(transition)
                    for metastable in metastables:

                        if not metastable >= 0:
                            raise ValueError('Donor metastable level cannot be less than zero.')

                        data = rates[donor][receiver][ionisation][transition][metastable]

                        # sanitise/validate data
                        data['qref'] = float(data['qref'])
                        sanitise_and_validate(data, 'eb', 'beam energy', 'qeb', 'beam energy effective rate')
                        sanitise_and_validate(data, 'ti', 'ion temperature', 'qti', 'ion temperature effective rate')
                        sanitise_and_validate(data, 'ni', 'ion density', 'qni', 'ion density effective rate')
                        sanitise_and_validate(data, 'z', 'Zeff', 'qz', 'Zeff effective rate')
                        sanitise_and_validate(data, 'b', 'B-field magnitude', 'qb', 'B-field magnitude effective rate')

                        content[transition_key][metastable] = {
                            'eb': data['eb'].tolist(),
                            'ti': data['ti'].tolist(),
                            'ni': data['ni'].tolist(),
                            'z': data['z'].tolist(),
                            'b': data['b'].tolist(),
                            'qref': data['qref'],
                            'qeb': data['qeb'].tolist(),
                            'qti': data['qti'].tolist(),
                            'qni': data['qni'].tolist(),
                            'qz': data['qz'].tolist(),
                            'qb': data['qb'].tolist(),
                        }

                # create directory structure if missing
                directory = os.path.dirname(path)
                if not os.path.isdir(directory):
                    os.makedirs(directory)

                # write new data
                with open(path, 'w') as f:
                    json.dump(content, f, indent=2, sort_keys=True)


def get_beam_cx_rates(donor_ion, receiver_ion, receiver_ionisation, transition, repository_path=None):

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH
    path = os.path.join(repository_path, 'beam/cx/{}/{}/{}.json'.format(donor_ion.symbol.lower(), receiver_ion.symbol.lower(), receiver_ionisation))
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        rates = content[encode_transition(transition)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested beam CX effective emission rates (donor={}, receiver={}, ionisation={}, transition={})'
                           ' are not available.'.format(donor_ion.symbol, receiver_ion.symbol, receiver_ionisation, transition))

    # sanitise data and convert to (more useful) numpy arrays rather than lists
    for rate in rates.values():
        rate['eb'] = np.array(rate['eb'], np.float64)
        rate['ti'] = np.array(rate['ti'], np.float64)
        rate['ni'] = np.array(rate['ni'], np.float64)
        rate['z'] = np.array(rate['z'], np.float64)
        rate['b'] = np.array(rate['b'], np.float64)
        rate['qref'] = float(rate['qref'])
        rate['qeb'] = np.array(rate['qeb'], np.float64)
        rate['qti'] = np.array(rate['qti'], np.float64)
        rate['qni'] = np.array(rate['qni'], np.float64)
        rate['qz'] = np.array(rate['qz'], np.float64)
        rate['qb'] = np.array(rate['qb'], np.float64)

    return [(int(metastable), rate) for metastable, rate in rates.items()]



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
import re

# todo: to be implemented in a future release


# def parse_adf11acd(ion, ionisation, adf_file_path):
#     pass


# def parse_adf11scd(ion, ionisation, adf_file_path):
#     pass


# def parse_adf11qcd(ion, ionisation, adf_file_path):
#     pass


# def parse_adf11xcd(ion, ionisation, adf_file_path):
#     pass


# def parse_adf11ccd(ion, ionisation, adf_file_path):
#     pass


# def parse_adf11plt(ion, ionisation, adf_file_path):
#     pass


# def parse_adf11prb(ion, ionisation, adf_file_path):
#     pass


# def parse_adf11pls(ion, ionisation, adf_file_path):
#     pass


# def parse_adf11prc(ion, ionisation, adf_file_path):
#     pass


# def parse_adf11xxx(file_path, ion, ionisation):
#
#     adf11_fh = open(file_path, 'r')
#     lines = adf11_fh.readlines()
#     adf11_fh.close()
#
#     file_header = lines.pop(0).split()
#     if int(file_header[0]) != ion.atomic_number:
#         raise ValueError("ADF file does not match required atomic number.")
#
#     num_densities = int(file_header[1])
#     num_temperatures = int(file_header[2])
#
#     lines.pop(0)
#     parameter_lines = []
#     while True:
#         next_line = lines.pop(0)
#         if next_line[0:2] == '--':
#             break
#         components = next_line.split()
#         for c in components:
#             parameter_lines.append(float(c))
#
#     if len(parameter_lines) != num_densities + num_temperatures:
#         raise ValueError("ADF11 file could not be parsed correctly.")

#     # todo: remove log10 remapping, interpolate directly! Note also needs to be fixed in rate.
#     densities = np.array([10**x * 1E4 for x in parameter_lines[0:num_densities]])
#     temperatures = np.array([10**x for x in parameter_lines[num_densities:]])
#
#     found = False
#     while True:
#         if next_line[0:2] == '--':
#             match = re.match('.*Z1= ([0-9]*).*', next_line)
#             if match and int(match.group(1)) == ionisation+1:
#                 found = True
#                 break
#         if next_line[0] == 'C':
#             break
#         next_line = lines.pop(0)
#
#     if found:
#         rate_data = []
#         while True:
#             next_line = lines.pop(0)
#             if next_line[0:2] == '--' or next_line[0] == 'C':
#                 break
#             for c in next_line.split():
#                 rate_data.append(10**float(c) * 1E-8)
#
#         rate_data = np.array(rate_data)
#         rate_data = rate_data.reshape((num_densities, num_temperatures))
#
#     else:
#         raise ValueError('Requested ADF11 data could not be found in this file.')
#
#     return densities, temperatures, rate_data



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from cherab.core import AtomicData
from cherab.core.atomic.elements import Isotope
from cherab.openadas.repository import DEFAULT_REPOSITORY_PATH

from .rates import *
from cherab.openadas import repository


class OpenADAS(AtomicData):
    """

    """

    def __init__(self, data_path=None, permit_extrapolation=False, missing_rates_return_null=False):

        super().__init__()
        self._data_path = data_path or DEFAULT_REPOSITORY_PATH

        # if true informs interpolation objects to allow extrapolation beyond the limits of the tabulated data
        self._permit_extrapolation = permit_extrapolation

        # if true, allows Null rate objects to be returned when the requested atomic data is missing
        self._missing_rates_return_null = missing_rates_return_null

    @property
    def data_path(self):
        return self._data_path

    def wavelength(self, ion, ionisation, transition):
        """
        :param ion: Element object defining the ion type.
        :param ionisation: Ionisation level of the ion.
        :param transition: Tuple containing (initial level, final level)
        :return: Wavelength in nanometers.
        """

        if isinstance(ion, Isotope):
            ion = ion.element
        return repository.get_wavelength(ion, ionisation, transition)

    def beam_cx_rate(self, donor_ion, receiver_ion, receiver_ionisation, transition):
        """

        :param donor_ion:
        :param receiver_ion:
        :param receiver_ionisation:
        :param transition:
        :return:
        """

        # extract element from isotope
        if isinstance(donor_ion, Isotope):
            donor_ion = donor_ion.element

        if isinstance(receiver_ion, Isotope):
            receiver_ion = receiver_ion.element

        try:
            # read data
            wavelength = repository.get_wavelength(receiver_ion, receiver_ionisation - 1, transition)
            data = repository.get_beam_cx_rates(donor_ion, receiver_ion, receiver_ionisation, transition)

        except RuntimeError:
            if self._missing_rates_return_null:
                return [NullBeamCXRate()]
            raise

        # load and interpolate the relevant transition data from each file
        rates = []
        for donor_metastable, rate_data in data:
            rates.append(BeamCXRate(donor_metastable, wavelength, rate_data, extrapolate=self._permit_extrapolation))
        return rates

    def beam_stopping_rate(self, beam_ion, plasma_ion, ionisation):
        """

        :param beam_ion:
        :param plasma_ion:
        :param ionisation:
        :return:
        """

        # extract element from isotope
        if isinstance(beam_ion, Isotope):
            beam_ion = beam_ion.element

        if isinstance(plasma_ion, Isotope):
            plasma_ion = plasma_ion.element

        try:
            # locate data file
            data = repository.get_beam_stopping_rate(beam_ion, plasma_ion, ionisation)

        except RuntimeError:
            if self._missing_rates_return_null:
                return NullBeamStoppingRate()
            raise

        # load and interpolate data
        return BeamStoppingRate(data, extrapolate=self._permit_extrapolation)

    def beam_population_rate(self, beam_ion, metastable, plasma_ion, ionisation):
        """

        :param beam_ion:
        :param metastable:
        :param plasma_ion:
        :param ionisation:
        :return:
        """

        # extract element from isotope
        if isinstance(beam_ion, Isotope):
            beam_ion = beam_ion.element

        if isinstance(plasma_ion, Isotope):
            plasma_ion = plasma_ion.element

        try:
            # locate data file
            data = repository.get_beam_population_rate(beam_ion, metastable, plasma_ion, ionisation)

        except RuntimeError:
            if self._missing_rates_return_null:
                return NullBeamPopulationRate()
            raise

        # load and interpolate data
        return BeamPopulationRate(data, extrapolate=self._permit_extrapolation)

    def beam_emission_rate(self, beam_ion, plasma_ion, ionisation, transition):
        """

        :param beam_ion:
        :param plasma_ion:
        :param ionisation:
        :param transition:
        :return:
        """

        # extract element from isotope
        if isinstance(beam_ion, Isotope):
            beam_ion = beam_ion.element

        if isinstance(plasma_ion, Isotope):
            plasma_ion = plasma_ion.element

        try:
            # locate data file
            data = repository.get_beam_emission_rate(beam_ion, plasma_ion, ionisation, transition)
            wavelength = repository.get_wavelength(plasma_ion, ionisation - 1, transition)

        except RuntimeError:
            if self._missing_rates_return_null:
                return NullBeamEmissionRate()
            raise

        # load and interpolate data
        return BeamEmissionRate(data, wavelength, extrapolate=self._permit_extrapolation)

    def impact_excitation_rate(self, ion, ionisation, transition):
        """

        :param ion:
        :param ionisation:
        :param transition:
        :return:
        """

        if isinstance(ion, Isotope):
            ion = ion.element

        try:
            wavelength = repository.get_wavelength(ion, ionisation, transition)
            data = repository.get_pec_excitation_rate(ion, ionisation, transition)

        except RuntimeError:
            if self._missing_rates_return_null:
                return NullImpactExcitationRate()
            raise

        return ImpactExcitationRate(wavelength, data, extrapolate=self._permit_extrapolation)

    def recombination_rate(self, ion, ionisation, transition):
        """

        :param ion:
        :param ionisation:
        :param transition:
        :return:
        """

        if isinstance(ion, Isotope):
            ion = ion.element

        try:
            wavelength = repository.get_wavelength(ion, ionisation, transition)
            data = repository.get_pec_recombination_rate(ion, ionisation, transition)

        except (FileNotFoundError, KeyError):
            if self._missing_rates_return_null:
                return NullRecombinationRate()
            raise

        return RecombinationRate(wavelength, data, extrapolate=self._permit_extrapolation)

    # def stage_resolved_line_ radiation_rate(self, ion, ionisation):
    #
    #     # extract element from isotope
    #     if isinstance(ion, Isotope):
    #         ion = ion.element
    #
    #     name = 'Stage Resolved Line Radiation - ({}, {})'.format(ion.symbol, ionisation)
    #     return StageResolvedRadiation(ion, ionisation, densities, temperatures, rate_data,
    #                                   name=name, extrapolate=self._permit_extrapolation)

    # def stage_resolved_continuum_radiation_rate(self, ion, ionisation):
    #
    #     # extract element from isotope
    #     if isinstance(ion, Isotope):
    #         ion = ion.element
    #
    #     name = 'Stage Resolved Continuum Radiation - ({}, {})'.format(ion.symbol, ionisation)
    #
    #     return StageResolvedRadiation(ion, ionisation, densities, temperatures, rate_data,
    #                                   name=name, extrapolate=self._permit_extrapolation)





# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
cimport numpy as np
import matplotlib.pyplot as plt

from cherab.core.utility.conversion import PhotonToJ


cdef class ImpactExcitationRate(CoreImpactExcitationRate):

    def __init__(self, double wavelength, dict data, extrapolate=False):
        """
        :param wavelength: Resting wavelength of corresponding emission line in nm.
        :param data: Dictionary containing rate data.
        :param extrapolate: Enable extrapolation (default=False).
        """

        self.wavelength = wavelength
        self.raw_data = data

        # unpack
        ne = data['ne']
        te = data['te']
        rate =  data['rate']

        # pre-convert data to W m^3 from Photons s^-1 cm^3 prior to interpolation
        rate = PhotonToJ.to(rate, wavelength)

        # store limits of data
        self.density_range = ne.min(), ne.max()
        self.temperature_range = te.min(), te.max()

        # interpolate rate
        self._rate = Interpolate2DCubic(
            ne, te, rate, extrapolate=extrapolate, extrapolation_type="quadratic"
        )

    cpdef double evaluate(self, double density, double temperature) except? -1e999:

        # prevent -ve values (possible if extrapolation enabled)
        return max(0, self._rate.evaluate(density, temperature))


cdef class NullImpactExcitationRate(CoreImpactExcitationRate):
    """
    A PEC rate that always returns zero.
    Needed for use cases where the required atomic data is missing.
    """

    cpdef double evaluate(self, double density, double temperature) except? -1e999:
        return 0.0


cdef class RecombinationRate(CoreRecombinationRate):

    def __init__(self, double wavelength, dict data, extrapolate=False):
        """
        :param wavelength: Resting wavelength of corresponding emission line in nm.
        :param data: Dictionary containing rate data.
        :param extrapolate: Enable extrapolation (default=False).
        """

        self.wavelength = wavelength
        self.raw_data = data

        # unpack
        ne = data['ne']
        te = data['te']
        rate =  data['rate']

        # pre-convert data to W m^3 from Photons s^-1 cm^3 prior to interpolation
        rate = PhotonToJ.to(rate, wavelength)

        # store limits of data
        self.density_range = ne.min(), ne.max()
        self.temperature_range = te.min(), te.max()

        # interpolate rate
        self._rate = Interpolate2DCubic(
            ne, te, rate, extrapolate=extrapolate, extrapolation_type="quadratic"
        )


    cpdef double evaluate(self, double density, double temperature) except? -1e999:

        # prevent -ve values (possible if extrapolation enabled)
        return max(0, self._rate.evaluate(density, temperature))


cdef class NullRecombinationRate(CoreRecombinationRate):
    """
    A PEC rate that always returns zero.
    Needed for use cases where the required atomic data is missing.
    """

    cpdef double evaluate(self, double density, double temperature) except? -1e999:
        return 0.0



# cdef class ThermalCXRate(CoreThermalCXRate):
#
#     def __init__(self, double wavelength ,dict rate_data, extrapolate=False):
#         pass
#     cpdef double evaluate(self, double density, double temperature):
#         pass
#
#     def plot(self, density_list=None, x_limit=None, y_limit=None):
#
#         plt.figure()
#         for dens in density_list:
#             rates = [self.__call__(dens, temp) for temp in self._temperature]
#             plt.loglog(self._temperature, rates, label='{:.4G} m$^{{-3}}$'.format(dens))
#
#         if x_limit:
#             plt.xlim(x_limit)
#         if y_limit:
#             plt.ylim(y_limit)
#
#         plt.legend(loc=4)
#         plt.xlabel('Electron Temperature (eV)')
#         plt.ylabel('$PEC$ (m$^3$ s$^{-1}$)')
#         plt.title('Photon emissivity coefficient')
#
#
# cdef class NullThermalCXRate(CoreThermalCXRate):
#     """
#     A PEC rate that always returns zero.
#     Needed for use cases where the required atomic data is missing.
#     """
#
#     cpdef double evaluate(self, double density, double temperature) except? -1e999:
#         return 0.0



from setuptools import setup, find_packages, Extension
import sys
import numpy
import os
import os.path as path
import multiprocessing

use_cython = True
force = False
profile = False
install_rates = True

if "--skip-cython" in sys.argv:
    use_cython = False
    del sys.argv[sys.argv.index("--skip-cython")]

if "--force" in sys.argv:
    force = True
    del sys.argv[sys.argv.index("--force")]

if "--profile" in sys.argv:
    profile = True
    del sys.argv[sys.argv.index("--profile")]

if "--skip-rates-install" in sys.argv:
    install_rates = False
    del sys.argv[sys.argv.index("--skip-rates-install")]

source_paths = ['cherab', 'demos']
compilation_includes = [".", numpy.get_include()]
compilation_args = []
cython_directives = {
    'language_level': 3
}
setup_path = path.dirname(path.abspath(__file__))

if use_cython:

    from Cython.Build import cythonize

    # build .pyx extension list
    extensions = []
    for package in source_paths:
        for root, dirs, files in os.walk(path.join(setup_path, package)):
            for file in files:
                if path.splitext(file)[1] == ".pyx":
                    pyx_file = path.relpath(path.join(root, file), setup_path)
                    module = path.splitext(pyx_file)[0].replace("/", ".")
                    extensions.append(Extension(module, [pyx_file], include_dirs=compilation_includes, extra_compile_args=compilation_args),)

    if profile:
        cython_directives["profile"] = True

    # generate .c files from .pyx
    extensions = cythonize(extensions, nthreads=multiprocessing.cpu_count(), force=force, compiler_directives=cython_directives)

else:

    # build .c extension list
    extensions = []
    for package in source_paths:
        for root, dirs, files in os.walk(path.join(setup_path, package)):
            for file in files:
                if path.splitext(file)[1] == ".c":
                    c_file = path.relpath(path.join(root, file), setup_path)
                    module = path.splitext(c_file)[0].replace("/", ".")
                    extensions.append(Extension(module, [c_file], include_dirs=compilation_includes, extra_compile_args=compilation_args),)

# parse the package version number
with open(path.join(path.dirname(__file__), 'cherab/openadas/VERSION')) as version_file:
    version = version_file.read().strip()

setup(
    name="cherab-openadas",
    version=version,
    license="EUPL 1.1",
    namespace_packages=['cherab'],
    description='Cherab spectroscopy framework: OpenADAS atomic data source package',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics"
    ],
    install_requires=['cherab', 'numpy', 'cython>=0.28'],
    packages=find_packages(),
    include_package_data=True,
    zip_safe= False,
    ext_modules=extensions
)

# setup a rate repository with common rates
if install_rates:
    try:
        from cherab.openadas import repository
        repository.populate()
    except ImportError:
        pass




# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from cherab.core.utility.conversion import Cm3ToM3, PerCm3ToPerM3, PhotonToJ

cimport cython
from cherab.core.math cimport Interpolate1DCubic


cdef class BeamCXRate(CoreBeamCXRate):
    """
    The effective cx rate interpolation class.

    :param donor_metastable: The metastable state of the donor species for which the rate data applies.
    :param wavelength: The natural wavelength of the emission line associated with the rate data in nm.
    :param data: A dictionary holding the rate data.
    :param extrapolate: Set to True to enable extrapolation, False to disable (default).
    """

    @cython.cdivision(True)
    def __init__(self, int donor_metastable, double wavelength, dict data, bint extrapolate=False):

        self.donor_metastable = donor_metastable
        self.wavelength = wavelength
        self.raw_data = data

        # pre-convert data to W m^3 from Photons s^-1 m^3 prior to interpolation
        eb = data["eb"]                                          # eV/amu
        ti = data["ti"]                                          # eV
        ni = data["ni"]                                          # m^-3
        zeff = data["z"]                                         # dimensionless
        bmag = data["b"]                                         # Tesla

        qref = data["qref"]                                      # m^3/s
        qeb = PhotonToJ.to(data["qeb"], wavelength)              # W.m^3
        qti = data["qti"] / qref                                 # dimensionless
        qni = data["qni"] / qref                                 # dimensionless
        qzeff = data["qz"] / qref                                # dimensionless
        qbmag = data["qb"] / qref                                # dimensionless

        # store limits of data
        self.beam_energy_range = eb.min(), eb.max()
        self.density_range = ni.min(), ni.max()
        self.temperature_range = ti.min(), ti.max()
        self.zeff_range = zeff.min(), zeff.max()
        self.b_field_range = bmag.min(), bmag.max()

        # interpolate the rate data
        self._eb = Interpolate1DCubic(eb, qeb, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")
        self._ti = Interpolate1DCubic(ti, qti, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")
        self._ni = Interpolate1DCubic(ni, qni, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")
        self._zeff = Interpolate1DCubic(zeff, qzeff, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")
        self._b = Interpolate1DCubic(bmag, qbmag, tolerate_single_value=True, extrapolate=extrapolate, extrapolation_type="quadratic")

    cpdef double evaluate(self, double energy, double temperature, double density, double z_effective, double b_field) except? -1e999:
        """
        Interpolates and returns the effective cx rate for the given plasma parameters.

        If the requested data is out-of-range then the call with throw a ValueError exception.

        :param energy: Interaction energy in eV/amu.
        :param temperature: Receiver ion temperature in eV.
        :param density: Receiver ion density in m^-3
        :param z_effective: Plasma Z-effective.
        :param b_field: Magnetic field magnitude in Tesla.
        :return: The effective cx rate in W.m^3
        """

        cdef double rate

        rate = self._eb.evaluate(energy)
        if rate <= 0:
            return 0.0

        rate *= self._ti.evaluate(temperature)
        if rate <= 0:
            return 0.0

        rate *= self._ni.evaluate(density)
        if rate <= 0:
            return 0.0

        rate *= self._zeff.evaluate(z_effective)
        if rate <= 0:
            return 0.0

        rate *= self._b.evaluate(b_field)
        if rate <= 0:
            return 0.0

        return rate


cdef class NullBeamCXRate(CoreBeamCXRate):
    """
    A beam CX rate that always returns zero.
    Needed for use cases where the required atomic data is missing.
    """

    cpdef double evaluate(self, double energy, double temperature, double density, double z_effective, double b_field) except? -1e999:
        return 0.0
